import copy

from xcffib.xproto import StackMode
from libqtile.layout.base import Layout

from .node import Node, HORIZONTAL, VERTICAL


def convert_names(tree):
    return [convert_names(n) if isinstance(n, list) else n.payload.name
            for n in tree]

class Plasma(Layout):

    defaults = [
        ('name', 'Plasma', 'Layout name'),
        ('border_normal', '#333333', 'Unfocused window border color'),
        ('border_focus', '#00e891', 'Focused window border color'),
        ('border_normal_fixed', '#333333',
         'Unfocused fixed-size window''border color'),
        ('border_focus_fixed', '#00e8dc',
         'Focused fixed-size window border color'),
        ('border_width', 1, 'Border width'),
        ('border_width_single', 0, 'Border width for single window'),
        ('margin', 0, 'Layout margin'),
    ]

    def __init__(self, **config):
        super().__init__(**config)
        self.add_defaults(Plasma.defaults)
        self.root = Node()
        self.focused = None
        self.split_orient = None

    def info(self):
        info = super().info()
        info['tree'] = convert_names(self.root.tree)
        return info

    def clone(self, group):
        clone = copy.copy(self)
        clone.group = group
        clone.root = Node()
        clone.focused = None
        clone.split_orient = None
        return clone

    def add(self, client):
        node = Node(client)
        if self.focused is None or self.focused_node is None:
            self.root.add_child(node)
            return
        if self.split_orient in (None, self.focused_node.parent.orient):
            self.focused_node.parent.add_child_after(node, self.focused_node)
        else:
            self.focused_node.split_with(node)
        self.split_orient = None

    def remove(self, client):
        self.root.find_payload(client).remove()

    def configure(self, client, screen):
        self.root.x = screen.x
        self.root.y = screen.y
        self.root.width = screen.width
        self.root.height = screen.height
        node = self.root.find_payload(client)
        border_width = \
            self.border_width_single if self.root.tree == [node] else \
            self.border_width
        border_color = getattr(self, 'border_' +
                               ('focus' if client.has_focus else 'normal') +
                               ('' if node.flexible else '_fixed'))
        x, y, width, height = node.pixel_perfect
        client.place(
            x,
            y,
            width-2*border_width,
            height-2*border_width,
            border_width,
            self.group.qtile.colorPixel(border_color),
            margin=self.margin,
        )
        # Always keep tiles below floating windows
        client.window.configure(stackmode=StackMode.Below)
        client.unhide()

    def focus(self, client):
        self.focused = client
        self.root.find_payload(client).access()

    def focus_first(self):
        return self.root.first_leaf.payload

    def focus_last(self):
        return self.root.last_leaf.payload

    def focus_next(self, win):
        next_leaf = self.root.find_payload(win).next_leaf
        return None if next_leaf is self.root.first_leaf else next_leaf.payload

    def focus_previous(self, win):
        prev_leaf = self.root.find_payload(win).prev_leaf
        return None if prev_leaf is self.root.last_leaf else prev_leaf.payload

    def focus_node(self, node):
        if node is None:
            return
        self.group.focus(node.payload)

    @property
    def focused_node(self):
        return self.root.find_payload(self.focused)

    def refocus(self):
        self.group.focus(self.focused)

    def cmd_next(self):
        """Focus next window."""
        self.focus_node(self.focused_node.next_leaf)

    def cmd_previous(self):
        """Focus previous window."""
        self.focus_node(self.focused_node.prev_leaf)

    def cmd_left(self):
        """Focus window to the left."""
        self.focus_node(self.focused_node.close_left)

    def cmd_right(self):
        """Focus window to the right."""
        self.focus_node(self.focused_node.close_right)

    def cmd_up(self):
        """Focus window above."""
        self.focus_node(self.focused_node.close_up)

    def cmd_down(self):
        """Focus window below."""
        self.focus_node(self.focused_node.close_down)

    def cmd_move_left(self):
        """Move current window left."""
        self.focused_node.move_left()
        self.refocus()

    def cmd_move_right(self):
        """Move current window right."""
        self.focused_node.move_right()
        self.refocus()

    def cmd_move_up(self):
        """Move current window up."""
        self.focused_node.move_up()
        self.refocus()

    def cmd_move_down(self):
        """Move current window down."""
        self.focused_node.move_down()
        self.refocus()

    def cmd_integrate_left(self):
        """Integrate current window left."""
        self.focused_node.integrate_left()
        self.refocus()

    def cmd_integrate_right(self):
        """Integrate current window right."""
        self.focused_node.integrate_right()
        self.refocus()

    def cmd_integrate_up(self):
        """Integrate current window up."""
        self.focused_node.integrate_up()
        self.refocus()

    def cmd_integrate_down(self):
        """Integrate current window down."""
        self.focused_node.integrate_down()
        self.refocus()

    def cmd_split_horizontal(self):
        """Add next window horizontally."""
        self.split_orient = HORIZONTAL

    def cmd_split_vertical(self):
        """Add next window vertically."""
        self.split_orient = VERTICAL

    def cmd_size(self, val):
        """Change size of current window.

        (It's recommended to use `width()`/`height()` instead.)
        """
        self.focused_node.size = val
        self.refocus()

    def cmd_width(self, val):
        """Set width of current window."""
        self.focused_node.width = val
        self.refocus()

    def cmd_height(self, val):
        """Set height of current window."""
        self.focused_node.height = val
        self.refocus()

    def cmd_reset_size(self):
        """Reset size of current window to automatic (relative) sizing."""
        self.focused_node.reset_size()
        self.refocus()

    def cmd_grow(self, amt):
        """Grow size of current window.

        (It's recommended to use `grow_width()`/`grow_height()` instead.)
        """
        self.focused_node.size += amt
        self.refocus()

    def cmd_grow_width(self, amt):
        """Grow width of current window."""
        self.focused_node.width += amt
        self.refocus()

    def cmd_grow_height(self, amt):
        """Grow height of current window."""
        self.focused_node.height += amt
        self.refocus()



from collections import defaultdict


class Canvas:

    horizontal_line = '\u2500'
    vertical_line = '\u2502'
    tl_corner = '\u250c'
    tr_corner = '\u2510'
    bl_corner = '\u2514'
    br_corner = '\u2518'

    def __init__(self, width, height):
        self.width = width
        self.height = height
        self.canvas = defaultdict(lambda: defaultdict(lambda: '#'))

    def add_box(self, x, y, width, height, name='*'):
        width = width-1
        height = height-1
        label = str(name)[:width-1]
        for i in range(x, width+x):
            self.canvas[i][y] = self.horizontal_line
            self.canvas[i][y+height] = self.horizontal_line
        for i in range(y, height+y):
            self.canvas[x][i] = self.vertical_line
            self.canvas[x+width][i] = self.vertical_line
        for i in range(x+1, width+x):
            for j in range(y+1, y+height):
                self.canvas[i][j] = '.'
        for i, char in enumerate(label):
            self.canvas[x+1+i][y+1] = char
        self.canvas[x][y] = self.tl_corner
        self.canvas[x+width][y] = self.tr_corner
        self.canvas[x][y+height] = self.bl_corner
        self.canvas[x+width][y+height] = self.br_corner

    def view(self):
        res = ''
        for y in range(self.height):
            for x in range(self.width):
                res += self.canvas[x][y]
            res += '\n'
        return res

def tree(node, level=0):
    res = '{indent}{name} {orient} {repr_} {pos} {size} {parent}\n'.format(
        indent=level*4*' ',
        name='%s' % (node.payload or '*'),
        orient='H' if node.horizontal else 'V',
        repr_='%s' % repr(node),
        pos='%g*%g@%g:%g' % (node.width, node.height, node.x, node.y),
        size='size: %s%s' % (node.size, ' (auto)' if node.flexible else ''),
        parent='p: %s' % node.parent,
    )
    for child in node.children:
        res += tree(child, level+1)
    return res

def draw(root):
    canvas = Canvas(root.width, root.height)
    def add(node):
        if node.is_leaf:
            canvas.add_box(
                *node.pixel_perfect,
                node.payload
            )
        for child in node.children:
            add(child)
    add(root)
    return canvas.view()

def info(node):
    print(tree(node))
    print(draw(node))



from collections import namedtuple
from datetime import datetime
from enum import Enum, auto
from math import isclose


HORIZONTAL = False
VERTICAL = True
Dimensions = namedtuple('Dimensions', 'x y width height')
Point = namedtuple('Point', 'x y')

class Direction(Enum):

    UP = auto()
    DOWN = auto()
    LEFT = auto()
    RIGHT = auto()

    @property
    def orientation(self):
        return HORIZONTAL if self.name in ['LEFT', 'RIGHT'] else VERTICAL

    @property
    def offset(self):
        return 1 if self.name in ['RIGHT', 'DOWN'] else -1

UP, DOWN, LEFT, RIGHT = list(Direction)

border_check = {
    UP: lambda a, b: isclose(a.y, b.y_end),
    DOWN: lambda a, b: isclose(a.y_end, b.y),
    LEFT: lambda a, b: isclose(a.x, b.x_end),
    RIGHT: lambda a, b: isclose(a.x_end, b.x),
}

def fit_into(nodes, space):
    """Resize nodes to fit them into the specified space."""
    if not nodes:
        return
    occupied = sum(n.min_size for n in nodes)
    if space >= occupied and any(n.flexible for n in nodes):
        # If any flexible node exists, it will occupy the space automatically,
        # not requiring any action.
        return
    nodes_left = nodes.copy()
    space_left = space
    if space < occupied:
        for node in nodes:
            if node.min_size_bound != node.min_size:
                continue
            # Substract nodes that are already at their minimal possible size
            # because they can't be shrinked any further.
            space_left -= node.min_size
            nodes_left.remove(node)
    if not nodes_left:
        return
    factor = space_left / sum(n.size for n in nodes_left)
    for node in nodes_left:
        new_size = node.size * factor
        if node.fixed:
            node._size = new_size  # pylint: disable=protected-access
        for child in node.children:
            fit_into(child.children, new_size)

class Node:

    min_size_default = 100
    root_orient = HORIZONTAL

    def __init__(self, payload=None, x=None, y=None, width=None, height=None):
        self.payload = payload
        self._x = x
        self._y = y
        self._width = width
        self._height = height
        self._size = None
        self.children = []
        self.last_accessed = datetime.min
        self.parent = None

    def __repr__(self):
        info = self.payload or ''
        if self.children:
            info += ' +%d' % len(self.children)
        return '<Node %s %x>' % (info, id(self))

    @property
    def is_root(self):
        return self.parent is None

    @property
    def root(self):
        try:
            return self.parent.root
        except AttributeError:
            return self

    @property
    def index(self):
        return self.parent.children.index(self)

    @property
    def siblings(self):
        return [c for c in self.parent.children if c is not self]

    @property
    def is_leaf(self):
        return not self.children

    @property
    def first_leaf(self):
        if self.is_leaf:
            return self
        return self.children[0].first_leaf

    @property
    def last_leaf(self):
        if self.is_leaf:
            return self
        return self.children[-1].last_leaf

    @property
    def recent_leaf(self):
        if self.is_leaf:
            return self
        return max(self.children, key=lambda n: n.last_accessed).recent_leaf

    @property
    def prev_leaf(self):
        if self.is_root:
            return self.last_leaf
        idx = self.index - 1
        if idx < 0:
            return self.parent.prev_leaf
        return self.parent.children[idx].last_leaf

    @property
    def next_leaf(self):
        if self.is_root:
            return self.first_leaf
        idx = self.index + 1
        if idx >= len(self.parent.children):
            return self.parent.next_leaf
        return self.parent.children[idx].first_leaf

    @property
    def orient(self):
        if self.is_root:
            return self.root_orient
        return not self.parent.orient

    @property
    def horizontal(self):
        return self.orient is HORIZONTAL

    @property
    def vertical(self):
        return self.orient is VERTICAL

    @property
    def tree(self):
        return [c.tree if c.children else c for c in self.children]

    @property
    def all_leafs(self):
        if self.is_leaf:
            yield self
        for child in self.children:
            yield from child.all_leafs

    @property
    def size_offset(self):
        return sum(c.size for c in self.parent.children[:self.index])

    @property
    def x(self):
        if self.is_root:
            return self._x
        if self.horizontal:
            return self.parent.x
        return self.parent.x + self.size_offset

    @x.setter
    def x(self, val):
        if not self.is_root:
            return
        self._x = val

    @property
    def y(self):
        if self.is_root:
            return self._y
        if self.vertical:
            return self.parent.y
        return self.parent.y + self.size_offset

    @y.setter
    def y(self, val):
        if not self.is_root:
            return
        self._y = val

    @property
    def pos(self):
        return Point(self.x, self.y)

    @property
    def width(self):
        if self.is_root:
            return self._width
        if self.horizontal:
            return self.parent.width
        return self.size

    @width.setter
    def width(self, val):
        if self.is_root:
            self._width = val
        elif self.horizontal:
            self.parent.size = val
        else:
            self.size = val

    @property
    def height(self):
        if self.is_root:
            return self._height
        if self.vertical:
            return self.parent.height
        return self.size

    @height.setter
    def height(self, val):
        if self.is_root:
            self._height = val
        elif self.vertical:
            self.parent.size = val
        else:
            self.size = val

    @property
    def x_end(self):
        return self.x + self.width

    @property
    def y_end(self):
        return self.y + self.height

    @property
    def x_center(self):
        return self.x + self.width / 2

    @property
    def y_center(self):
        return self.y + self.height / 2

    @property
    def center(self):
        return Point(self.x_center, self.y_center)

    @property
    def top_left(self):
        return Point(self.x, self.y)

    @property
    def top_right(self):
        return Point(self.x + self.width, self.y)

    @property
    def bottom_left(self):
        return Point(self.x, self.y + self.height)

    @property
    def bottom_right(self):
        return Point(self.x + self.width, self.y + self.height)

    @property
    def pixel_perfect(self):
        """Return pixel-perfect int dimensions (x, y, width, height) which
        compensate for gaps in the layout grid caused by plain int conversion.
        """
        x, y, width, height = self.x, self.y, self.width, self.height
        threshold = 0.99999
        if (x - int(x)) + (width - int(width)) > threshold:
            width += 1
        if (y - int(y)) + (height - int(height)) > threshold:
            height += 1
        return Dimensions(*map(int, (x, y, width, height)))

    @property
    def capacity(self):
        return self.width if self.horizontal else self.height

    @property
    def size(self):
        if self.is_root:
            return None
        if self.fixed:
            return self._size
        if self.flexible:
            # Distribute space evenly among flexible nodes
            taken = sum(n.size for n in self.siblings if not n.flexible)
            flexibles = [n for n in self.parent.children if n.flexible]
            return (self.parent.capacity - taken) / len(flexibles)
        return max(sum(gc.min_size for gc in c.children)
                   for c in self.children)

    @size.setter
    def size(self, val):
        if self.is_root or not self.siblings:
            return
        occupied = sum(s.min_size_bound for s in self.siblings)
        val = max(min(val, self.parent.capacity - occupied),
                  self.min_size_bound)
        self.force_size(val)

    def force_size(self, val):
        fit_into(self.siblings, self.parent.capacity - val)
        if val != 0:
            if self.children:
                fit_into([self], val)
            self._size = val

    @property
    def flexible(self):
        """A node is flexible if its size isn't (explicitly or implictly)
        determined.
        """
        if self.fixed:
            return False
        return all((any(gc.flexible for gc in c.children) or c.is_leaf)
                   for c in self.children)

    @property
    def fixed(self):
        """A node is fixed if it has a specified size."""
        return self._size is not None

    @property
    def min_size(self):
        if self.fixed:
            return self._size
        if self.is_leaf:
            return self.min_size_default
        size = max(sum(gc.min_size for gc in c.children)
                   for c in self.children)
        return max(size, self.min_size_default)

    @property
    def min_size_bound(self):
        if self.is_leaf:
            return self.min_size_default
        return max(sum(gc.min_size_bound for gc in c.children) or
                   self.min_size_default for c in self.children)

    def reset_size(self):
        self._size = None

    def grow(self, amt, orient=None):
        # TODO Deprecate grow
        if self.is_root:
            return
        if orient is (not self.parent.orient):
            self.parent.grow(amt)
            return
        self.size += amt

    def access(self):
        self.last_accessed = datetime.now()
        try:
            self.parent.access()
        except AttributeError:
            pass

    def neighbor(self, direction):
        """Return adjacent leaf node in specified direction."""
        if self.is_root:
            return None
        if direction.orientation is self.parent.orient:
            target_idx = self.index + direction.offset
            if 0 <= target_idx < len(self.parent.children):
                return self.parent.children[target_idx].recent_leaf
            if self.parent.is_root:
                return None
            return self.parent.parent.neighbor(direction)
        return self.parent.neighbor(direction)

    @property
    def up(self):
        return self.neighbor(UP)

    @property
    def down(self):
        return self.neighbor(DOWN)

    @property
    def left(self):
        return self.neighbor(LEFT)

    @property
    def right(self):
        return self.neighbor(RIGHT)

    def common_border(self, node, direction):
        """Return whether a common border with given node in specified
        direction exists.
        """
        if not border_check[direction](self, node):
            return False
        if direction in [UP, DOWN]:
            detached = node.x >= self.x_end or node.x_end <= self.x
        else:
            detached = node.y >= self.y_end or node.y_end <= self.y
        return not detached

    def close_neighbor(self, direction):
        """Return visually adjacent leaf node in specified direction."""
        nodes = [n for n in self.root.all_leafs if
                 self.common_border(n, direction)]
        if not nodes:
            return None
        most_recent = max(nodes, key=lambda n: n.last_accessed)
        if most_recent.last_accessed > datetime.min:
            return most_recent
        if direction in [UP, DOWN]:
            match = lambda n: n.x <= self.x_center <= n.x_end
        else:
            match = lambda n: n.y <= self.y_center <= n.y_end
        return next(n for n in nodes if match(n))

    @property
    def close_up(self):
        return self.close_neighbor(UP)

    @property
    def close_down(self):
        return self.close_neighbor(DOWN)

    @property
    def close_left(self):
        return self.close_neighbor(LEFT)

    @property
    def close_right(self):
        return self.close_neighbor(RIGHT)

    def add_child(self, node, idx=None):
        # TODO Auto-split-with if required
        if idx is None:
            idx = len(self.children)
        self.children.insert(idx, node)
        node.parent = self
        if len(self.children) == 1:
            return
        total = self.capacity
        fit_into(node.siblings, total - (total / len(self.children)))

    def add_child_after(self, new, old):
        self.add_child(new, idx=self.children.index(old)+1)

    def remove_child(self, node):
        node.force_size(0)
        self.children.remove(node)
        if len(self.children) != 1:
            return
        if not self.is_root:
            # Collapse tree with a single child
            self.parent.replace_child(self, self.children[0])
        else:
            # A single child doesn't need an absolute size
            self.children[0].reset_size()

    def remove(self):
        self.parent.remove_child(self)

    def replace_child(self, old, new):
        self.children[self.children.index(old)] = new
        new.parent = self
        new._size = old._size  # pylint: disable=protected-access

    def split_with(self, node):
        container = Node()
        self.parent.replace_child(self, container)
        self.reset_size()
        for child in [self, node]:
            container.add_child(child)

    def move(self, direction):
        if self.is_root:
            return
        if direction.orientation is self.parent.orient:
            old_idx = self.index
            new_idx = old_idx + direction.offset
            if 0 <= new_idx < len(self.parent.children):
                ch = self.parent.children
                ch[old_idx], ch[new_idx] = ch[new_idx], ch[old_idx]
                return
            new_sibling = self.parent.parent
        else:
            new_sibling = self.parent
        try:
            new_parent = new_sibling.parent
            idx = new_parent.children.index(new_sibling)
        except AttributeError:
            return
        self.reset_size()
        self.parent.remove_child(self)
        new_parent.add_child(self, idx + (1 if direction.offset == 1 else 0))

    def move_up(self):
        self.move(UP)

    def move_down(self):
        self.move(DOWN)

    def move_right(self):
        self.move(RIGHT)

    def move_left(self):
        self.move(LEFT)

    def _move_and_integrate(self, direction):
        old_parent = self.parent
        self.move(direction)
        if self.parent is not old_parent:
            self.integrate(direction)

    def integrate(self, direction):
        if direction.orientation != self.parent.orient:
            self._move_and_integrate(direction)
            return
        target_idx = self.index + direction.offset
        if target_idx < 0 or target_idx >= len(self.parent.children):
            self._move_and_integrate(direction)
            return
        self.reset_size()
        target = self.parent.children[target_idx]
        self.parent.remove_child(self)
        if target.is_leaf:
            target.split_with(self)
        else:
            target.add_child(self)

    def integrate_up(self):
        self.integrate(UP)

    def integrate_down(self):
        self.integrate(DOWN)

    def integrate_left(self):
        self.integrate(LEFT)

    def integrate_right(self):
        self.integrate(RIGHT)

    def find_payload(self, payload):
        if self.payload is payload:
            return self
        for node in self.children:
            needle = node.find_payload(payload)
            if needle:
                return needle
